import os

from anasymod.generators.codegen import CodeGenerator
from anasymod.util import back2fwd, expand_searchpaths
from typing import Union

class ConfigFileObj(CodeGenerator):
    def __init__(self, files, config_path, name):
        super().__init__()
        self.files = None
        """ type(str) : mandatory setting; defines the path to sources. The path can be relative, absolute and 
            contain wildcards. """

        if isinstance(files, list):
            self.files = files
        elif isinstance(files, str):
            self.files = [files]
        else:
            raise TypeError(f"Type of config_paths variable provided to SubConfig class is not a list, is:{type(files)} instead.")

        self.config_path = config_path
        self.name = name

    def expand_paths(self):
        """
        Expand environment variables in provided list of paths.
        Check if path is absolute or relative, in case of a relative path, it will be expanded to an absolute path,
        whereas the folder of the config_file will be used to complete the path.
        """

        self.files = expand_searchpaths(paths=self.files, rel_path_reference=os.path.dirname(self.config_path))

class SubConfig(ConfigFileObj):
    def __init__(self, files: Union[list, str], name, config_path=None):
        super().__init__(files=files, config_path=config_path, name=name)

class Sources(ConfigFileObj):
    def __init__(self, files: list, fileset, config_path, name):
        super().__init__(files=files, config_path=config_path, name=name)
        self.fileset = fileset
        """ type(str) : Fileset, the source shall be associsted with. """


    def generate(self):
        pass

    def set_property(self, name, value, objects):
        self.writeln(' '.join(['set_property', '-name', name, '-value', value, '-objects', objects]))

class VerilogSource(Sources):
    """
    Container for source of type Verilog/SystemVerilog.

    :param files: Path to source file, could be relative/absolute and contain wildcards
    :type files: str

    """
    def __init__(self, files: Union[list, str], name, fileset=r"default", config_path=None, version=None):
        super().__init__(files=files, fileset=fileset, config_path=config_path, name=name)
        self.version = version
        """ type(str) : Verilog version, that shall be used when compiling sources. """

    def generate(self):
        self.text = self.files
        return self.dump()

class VerilogHeader(Sources):
    def __init__(self, files: Union[list, str], name, fileset=r"default", config_path=None):
        super().__init__(files=files, fileset=fileset, config_path=config_path, name=name)

    def set_header_files(self):
        file_list = '{ ' + ' '.join('"' + back2fwd(file) + '"' for file in self.files) + ' }'
        self.set_property('file_type', '{Verilog Header}', f'[get_files {file_list}]')

    def generate(self):
        self.dump()

class VHDLSource(Sources):
    def __init__(self, files: Union[list, str], name, library=None, fileset=r"default", config_path=None, version=None):
        super().__init__(files=files, fileset=fileset, config_path=config_path, name=name)
        self.library = library
        """ type(str) : Library, the source shall be associated with when compiling. """

        self.version = version
        """ type(str) : VHDL version, that shall be used when compiling sources. """


    def generate(self):
        self.dump()

class IncludeDir(Sources):
    def __init__(self, files: Union[list, str], name, fileset=r"default", config_path=None):
        super().__init__(files=files, fileset=fileset, config_path=config_path, name=name)

class EDIFFile(Sources):
    def __init__(self, files: Union[list, str], name, fileset=r"default", config_path=None):
        super().__init__(files=files, fileset=fileset, config_path=config_path, name=name)

class FirmwareFile(Sources):
    def __init__(self, files: Union[list, str], name, fileset=r"default", config_path=None):
        super().__init__(files=files, fileset=fileset, config_path=config_path, name=name)

class XCIFile(Sources):
    def __init__(self, files: Union[list, str], name, fileset=r"default", config_path=None):
        super().__init__(files=files, fileset=fileset, config_path=config_path, name=name)

class TCLFile(Sources):
    def __init__(self, files: Union[list, str], name, fileset=r"default", config_path=None):
        super().__init__(files=files, fileset=fileset, config_path=config_path, name=name)

class XDCFile(Sources):
    def __init__(self, files: Union[list, str], name, fileset=r"default", config_path=None, xdc_mode=None):
        # set defaults
        if xdc_mode is None:
            xdc_mode = 'read_xdc'

        # save custom settings
        self.xdc_mode = xdc_mode

        # call the super constructor
        super().__init__(files=files, fileset=fileset, config_path=config_path, name=name)

class MEMFile(Sources):
    def __init__(self, files: str, name, fileset=r"default", config_path=None):
        super().__init__(files=[files], fileset=fileset, config_path=config_path, name=name)

class BDFile(Sources):
    def __init__(self, files: str, name, fileset=r"default", config_path=None):
        super().__init__(files=[files], fileset=fileset, config_path=config_path, name=name)

class IPRepo(Sources):
    def __init__(self, files: str, name, fileset=r"default", config_path=None):
        super().__init__(files=[files], fileset=fileset, config_path=config_path, name=name)

class FunctionalModel(Sources):
    def __init__(self, files: str, name, fileset=r"default", config_path=None):
        super().__init__(files=[files], fileset=fileset, config_path=config_path, name=name)
        self.gen_files = None

    def set_gen_files_path(self, hdl_dir_root):
        """
        Set the result HDL path, where generated files can be found after generation was conducted.

        :param hdl_dir_root: Root directory for gen_files, this is usually set in emu config.
        """
        # TODO: Have the model generator declare what files should be included in "gen_files"
        # It is possible that not everything in the hdl_dir_root is an HDL source (e.g.,
        # temporary files generated during processing, memory files that are included, etc.)
        self.gen_files = [os.path.join(hdl_dir_root, self.fileset, self.name, '*.*v')]

    def expand_gen_files_path(self):
        """
        Expand environment variables in provided list of paths.
        Check if path is absolute or relative, in case of a relative path, it will be expanded to an absolute path,
        whereas the folder of the config_file will be used to complete the path.
        """

        self.gen_files = expand_searchpaths(paths=self.gen_files, rel_path_reference=os.path.dirname(self.config_path))


class CodeGenerator:
    def __init__(self, tab_string: str=None, line_ending='\n'):
        # save settings
        self.tab_string = tab_string if tab_string is not None else '    '
        self.line_ending = line_ending if line_ending is not None else '\n'

        # initialize variables
        self.tab_level = 0
        self.text = ''

    def write(self, line):
        self.text += line

    def writeln(self, line=''):
        self.write(self.tab_level * self.tab_string + line + self.line_ending)

    def reset(self):
        self.text = ''

    def use_templ(self, template):
        self.write(template.render())

    def write_to_file(self, filename):
        with open(filename, 'w') as f:
            f.write(self.text)
        self.reset()

    def read_from_file(self, filename):
        with open(filename, 'r') as f:
            self.text += f.read()

    def dump(self):
        return self.text

    def indent(self, quantity=1):
        self.tab_level += (1 * quantity)

    def dedent(self, quantity=1):
        self.tab_level -= (1 * quantity)
        assert self.tab_level >= 0




import traceback
import os
from multiprocessing import current_process


def statpro_update(feature, log=None):
    """

    :type feature: str
    :type log: logging.Logger
    """
    if not "INICIO_VERSION_HOME" in os.environ:
        return

    if current_process().name != "MainProcess":
        return

    try:
        import inicioenv
        import inicio

        msg = "Sending statpro information about feature: " + feature
        if log is not None:
            log.debug(msg)

        cfg = inicio.config_dict()
        inicioenv.statpro_update([feature], cfg)

    except ImportError:
        msg = "Could not import inicioenv and inicio!"
        if log is not None:
            log.warning(msg)
        else:
            print(msg)
        return

    except Exception:
        msg = "Exception occured while statpro update: " + traceback.format_exc()
        if log is not None:
            log.warning(msg)
        else:
            print(msg)
        return


class FEATURES:
    anasymod_import         = "anasymod_import"
    anasymod_sim            = "anasymod_sim_"
    anasymod_build_vivado   = "anasymod_build_vivado"
    anasymod_emulate_vivado = "anasymod_emulate_vivado"


if __name__ == '__main__':
    statpro_update(FEATURES.anasymod_import)



from anasymod.templates.templ import JinjaTempl
from anasymod.sim_ctrl.datatypes import DigitalSignal

class TemplZynqGPIO(JinjaTempl):
    def __init__(self, ps_vlnv='xilinx.com:ip:processing_system7:5.5',
                 gpio_vlnv='xilinx.com:ip:axi_gpio:2.0', design_name='zynq_gpio',
                 ultra_ps_vlnv='xilinx.com:ip:zynq_ultra_ps_e:3.3',
                 is_ultrascale=False):
        # call the super constructor
        super().__init__(trim_blocks=False, lstrip_blocks=False)

        # save settings
        self.design_name=design_name
        self.ps_vlnv = ps_vlnv
        self.gpio_vlnv = gpio_vlnv
        self.ultra_ps_vlnv = ultra_ps_vlnv
        self.is_ultrascale = is_ultrascale

        # save IO that are used by the emulator
        self.o_ctrl = DigitalSignal(name='o_ctrl', width=32, abspath=None)
        self.o_data = DigitalSignal(name='o_data', width=32, abspath=None)
        self.i_ctrl = DigitalSignal(name='i_ctrl', width=32, abspath=None)
        self.i_data = DigitalSignal(name='i_data', width=32, abspath=None)

        # generate text

    TEMPLATE_TEXT = '''\
############################
# Create the block diagram #
############################

# Initialize

create_bd_design "{{subst.design_name}}"

# Instantiate IPs

{% if subst.is_ultrascale %}
create_bd_cell -type ip -vlnv {{subst.ultra_ps_vlnv}} zynq_ultra_ps_e_0
{% else %}
create_bd_cell -type ip -vlnv {{subst.ps_vlnv}} processing_system7_0
{% endif %}
create_bd_cell -type ip -vlnv {{subst.gpio_vlnv}} axi_gpio_0
create_bd_cell -type ip -vlnv {{subst.gpio_vlnv}} axi_gpio_1

# Configure IPs

set_property \\
    -dict [list \\
        CONFIG.C_IS_DUAL {1} \\
        CONFIG.C_ALL_OUTPUTS {1} \\
        CONFIG.C_ALL_INPUTS_2 {1} \\
    ] \\
    [get_bd_cells axi_gpio_0]

set_property \\
    -dict [list \\
        CONFIG.C_IS_DUAL {1} \\
        CONFIG.C_ALL_OUTPUTS {1} \\
        CONFIG.C_ALL_OUTPUTS_2 {1} \\
    ] \\
    [get_bd_cells axi_gpio_1]

# Create ports

create_bd_port -dir O -from 31 -to 0 o_ctrl 
create_bd_port -dir I -from 31 -to 0 o_data
create_bd_port -dir O -from 31 -to 0 i_ctrl 
create_bd_port -dir O -from 31 -to 0 i_data

# Wire up IPs

{% if not subst.is_ultrascale %}
connect_bd_net \\
    [get_bd_pins processing_system7_0/FCLK_CLK0] \\
    [get_bd_pins processing_system7_0/M_AXI_GP0_ACLK]
{% endif %}

connect_bd_net \\
    [get_bd_ports o_ctrl] \\
    [get_bd_pins axi_gpio_0/gpio_io_o]

connect_bd_net \\
    [get_bd_ports o_data] \\
    [get_bd_pins axi_gpio_0/gpio2_io_i]

connect_bd_net \\
    [get_bd_ports i_ctrl] \\
    [get_bd_pins axi_gpio_1/gpio_io_o]

connect_bd_net \\
    [get_bd_ports i_data] \\
    [get_bd_pins axi_gpio_1/gpio2_io_o]

####################
# Apply automation #
####################   

# processing system automation

{% if subst.is_ultrascale %}
apply_bd_automation \\
    -rule xilinx.com:bd_rule:zynq_ultra_ps_e \\
    -config { \\
        apply_board_preset "1" \\
    } \\
    [get_bd_cells zynq_ultra_ps_e_0]
{% else %}
apply_bd_automation \\
    -rule xilinx.com:bd_rule:processing_system7 \\
    -config { \\
        make_external "FIXED_IO, DDR" \\
        apply_board_preset "1" \\
        Master "Disable" \\
        Slave "Disable" \\
    } \\
    [get_bd_cells processing_system7_0]
{% endif %}

# axi_gpio_0 automation

{% if subst.is_ultrascale %}
apply_bd_automation \\
    -rule xilinx.com:bd_rule:axi4 \\
    -config { \\
        Clk_master {Auto} \\
        Clk_slave {Auto} \\
        Clk_xbar {Auto} \\
        Master {/zynq_ultra_ps_e_0/M_AXI_HPM0_FPD} \\
        Slave {/axi_gpio_0/S_AXI} \\
        ddr_seg {Auto} \\
        intc_ip {New AXI Interconnect} \\
        master_apm {0}\\
    } \\
    [get_bd_intf_pins axi_gpio_0/S_AXI]
{% else %}
apply_bd_automation \\
    -rule xilinx.com:bd_rule:axi4 \\
    -config { \\
        Clk_master {/processing_system7_0/FCLK_CLK0 (100 MHz)} \\
        Clk_slave {Auto} \\
        Clk_xbar {Auto} \\
        Master {/processing_system7_0/M_AXI_GP0} \\
        Slave {/axi_gpio_0/S_AXI} \\
        intc_ip {New AXI Interconnect} \\
        master_apm {0}\\
    } \\
    [get_bd_intf_pins axi_gpio_0/S_AXI]
{% endif %}

# axi_gpio_1 automation

{% if subst.is_ultrascale %}
apply_bd_automation \\
    -rule xilinx.com:bd_rule:axi4 \\
    -config { \\
        Clk_master {Auto} \\
        Clk_slave {Auto} \\
        Clk_xbar {Auto} \\
        Master {/zynq_ultra_ps_e_0/M_AXI_HPM0_FPD} \\
        Slave {/axi_gpio_1/S_AXI} \\
        ddr_seg {Auto} \\
        intc_ip {New AXI Interconnect} \\
        master_apm {0}\\
    } \\
    [get_bd_intf_pins axi_gpio_1/S_AXI]
{% else %}
apply_bd_automation \\
    -rule xilinx.com:bd_rule:axi4 \\
    -config { \\
        Clk_master {/processing_system7_0/FCLK_CLK0 (100 MHz)} \\
        Clk_slave {Auto} \\
        Clk_xbar {Auto} \\
        Master {/processing_system7_0/M_AXI_GP0} \\
        Slave {/axi_gpio_1/S_AXI} \\
        intc_ip {New AXI Interconnect} \\
        master_apm {0}\\
    } \\
    [get_bd_intf_pins axi_gpio_1/S_AXI]
{% endif %}

# other automation

{% if subst.is_ultrascale %}
apply_bd_automation \\
    -rule xilinx.com:bd_rule:axi4 \\
    -config { \\
        Clk_master {Auto} \\
        Clk_slave {/zynq_ultra_ps_e_0/pl_clk0 (99 MHz)} \\
        Clk_xbar {/zynq_ultra_ps_e_0/pl_clk0 (99 MHz)} \\
        Master {/zynq_ultra_ps_e_0/M_AXI_HPM1_FPD} \\
        Slave {/axi_gpio_0/S_AXI} \\
        ddr_seg {Auto} \\
        intc_ip {/ps8_0_axi_periph} \\
        master_apm {0}\\
    } \\
    [get_bd_intf_pins zynq_ultra_ps_e_0/M_AXI_HPM1_FPD]
{% endif %}

validate_bd_design
'''

    # save external IOs that need to be wired to the top
    EXT_IOS = [
        DigitalSignal(name='DDR_addr', width=15, abspath=None),
        DigitalSignal(name='DDR_ba', width=3, abspath=None),
        DigitalSignal(name='DDR_cas_n', width=1, abspath=None),
        DigitalSignal(name='DDR_ck_n', width=1, abspath=None),
        DigitalSignal(name='DDR_ck_p', width=1, abspath=None),
        DigitalSignal(name='DDR_cke', width=1, abspath=None),
        DigitalSignal(name='DDR_cs_n', width=1, abspath=None),
        DigitalSignal(name='DDR_dm', width=4, abspath=None),
        DigitalSignal(name='DDR_dq', width=32, abspath=None),
        DigitalSignal(name='DDR_dqs_n', width=4, abspath=None),
        DigitalSignal(name='DDR_dqs_p', width=4, abspath=None),
        DigitalSignal(name='DDR_odt', width=1, abspath=None),
        DigitalSignal(name='DDR_ras_n', width=1, abspath=None),
        DigitalSignal(name='DDR_reset_n', width=1, abspath=None),
        DigitalSignal(name='DDR_we_n', width=1, abspath=None),
        DigitalSignal(name='FIXED_IO_ddr_vrn', width=1, abspath=None),
        DigitalSignal(name='FIXED_IO_ddr_vrp', width=1, abspath=None),
        DigitalSignal(name='FIXED_IO_mio', width=54, abspath=None),
        DigitalSignal(name='FIXED_IO_ps_clk', width=1, abspath=None),
        DigitalSignal(name='FIXED_IO_ps_porb', width=1, abspath=None),
        DigitalSignal(name='FIXED_IO_ps_srstb', width=1, abspath=None)
    ]


class FirmwareGPIO:
    def __init__(self, scfg):
        crtl_inputs = scfg.analog_ctrl_inputs + scfg.digital_ctrl_inputs
        ctrl_outputs = scfg.analog_ctrl_outputs + scfg.digital_ctrl_outputs

        self.getter_dict = {}
        for probe in ctrl_outputs:
            if probe.o_addr is not None:
                self.getter_dict[probe.o_addr] = probe.name

        self.setter_dict = {}
        for param in crtl_inputs:
            if param.i_addr is not None:
                self.setter_dict[param.i_addr] = param.name

        self.hdr_text = self.gen_hdr_text()
        self.src_text = self.gen_src_text()

    def gen_hdr_text(self):
        retval = '''
#include "xgpio.h"

int init_GPIO();
'''

        # add setter declarations
        for k, v in self.setter_dict.items():
            retval += f'''
void set_{v}(u32 val);
'''

        # add getter declarations
        for k, v in self.getter_dict.items():
            retval += f'''
u32 get_{v}();
'''

        # return the code
        return retval

    def gen_src_text(self):
        retval = '''
#include "xparameters.h"
#include "xgpio.h"
#include "sleep.h"

XGpio Gpio0; // chan 1: o_ctrl, chan 2: o_data
XGpio Gpio1; // chan 1: i_ctrl, chan 2: i_data

int init_GPIO(){
    int status0, status1;
    status0 = XGpio_Initialize(&Gpio0, XPAR_GPIO_0_DEVICE_ID);
    status1 = XGpio_Initialize(&Gpio1, XPAR_GPIO_1_DEVICE_ID);
    if ((status0 == XST_SUCCESS) && (status1 == XST_SUCCESS)) {
        return 0;
    } else {
        return 1;
    }
}

u32 get_value(u32 addr){
    XGpio_DiscreteWrite(&Gpio0, 1, addr);
    usleep(1);
    return XGpio_DiscreteRead(&Gpio0, 2);
}

void set_value(u32 addr, u32 val){
    // set address and data
    XGpio_DiscreteWrite(&Gpio1, 1, addr);
    XGpio_DiscreteWrite(&Gpio1, 2, val);
    usleep(1);

    // assert "valid"
    XGpio_DiscreteWrite(&Gpio1, 1, addr | (1UL << 30));
    usleep(1);

    // de-assert "valid"
    XGpio_DiscreteWrite(&Gpio1, 1, addr);
    usleep(1);
}
'''

        # add setter definitions
        for k, v in self.setter_dict.items():
            retval += f'''
void set_{v}(u32 val){{
    set_value({k}, val);
}}
'''

        # add setter definitions
        for k, v in self.getter_dict.items():
            retval += f'''
u32 get_{v}(){{
    return get_value({k});
}}
'''
        # return the code
        return retval


import os.path
import subprocess

from anasymod.generators.vivado import VivadoTCLGenerator
from anasymod.generators.codegen import CodeGenerator
from anasymod.util import back2fwd

from anasymod.templates.dbg_hub import TemplDbgHub
from anasymod.templates.ext_clk import TemplExtClk
from anasymod.templates.clk_wiz import TemplClkWiz
from anasymod.templates.execute_FPGA_sim import TemplEXECUTE_FPGA_SIM
from anasymod.templates.ila import TemplILA
from anasymod.templates.zynq_gpio import TemplZynqGPIO
from anasymod.targets import FPGATarget
from anasymod.enums import FPGASimCtrl
from anasymod.sim_ctrl.ctrlapi import CtrlApi

class VivadoEmulation(VivadoTCLGenerator):
    """
    Generate and execute Vivado TCL scripts to generate a bitstream, run an emulation of FPGA for non-interactive mode,
    or launch an FPGA emulation for interactive mode and pass the handle for interactive control.
    """

    def __init__(self, target: FPGATarget):
        super().__init__(target=target)

    def build(self):
        #subst drive
        drive = 'V:'
        # Check if on-chip memory is sufficient on selected FPGA board
        if self.target.prj_cfg.board.bram is not None:
            bits_per_sample = 0
            probes = (self.target.str_cfg.digital_probes + self.target.str_cfg.analog_probes +
                      [self.target.str_cfg.time_probe] + [self.target.str_cfg.dec_cmp])
            for probe in probes:
                bits_per_sample += int(probe.width)
            if (bits_per_sample * self.target.prj_cfg.ila_depth) > (self.target.prj_cfg.board.bram * 8):
                raise(f'ERROR: Number ob samples to be recorded does not fit on FPGA board, please either select a '
                      f'board with more block memory, or change the ila_depth')
        else:
            print(f'WARNING: Check for sufficient BRAM could not be conducted, '
                  f'not enough information given in board definition!')

        scfg = self.target.str_cfg
        """ type : StructureConfig """

        project_root = self.target.project_root
        # under Windows there is the problem with path length more than 146 characters, that's why we have to use
        # subst command to substitute project directory to a drive letter
        if os.name == 'nt':
            if len(back2fwd(self.target.project_root)) > 80:

                project_root = self.subst_path(drive=drive)

        # create a new project
        self.create_project(
            project_name=self.target.prj_cfg.vivado_config.project_name,
            project_directory=project_root,
            full_part_name=self.target.prj_cfg.board.full_part_name,
            board_part=self.target.prj_cfg.board.board_part,
            force=True
        )

        # add all source files to the project (including header files)
        self.add_project_sources(content=self.target.content)

        # define the top module
        self.set_property('top', f"{{{self.target.cfg.top_module}}}", '[current_fileset]')

        # set define variables
        self.add_project_defines(content=self.target.content, fileset='[current_fileset]')

        # add include directories
        self.add_include_dirs(content=self.target.content, objects='[current_fileset]')

        # if desired, treat Verilog (*.v) files as SystemVerilog (*.sv)
        if self.target.prj_cfg.cfg.treat_v_as_sv:
            self.writeln('set_property file_type SystemVerilog [get_files -filter {FILE_TYPE == Verilog}]')

        # specify the level of flattening to use
        self.set_property(
            'STEPS.SYNTH_DESIGN.ARGS.FLATTEN_HIERARCHY',
            self.target.prj_cfg.cfg.flatten_hierarchy,
            '[get_runs synth_1]'
        )

        # append user constraints.  for flexibility and backwards compatibility, three modes are provided:
        # 1. "read_xdc": XDC file is read in using the "read_xdc" command (default)
        # 2. "pre_constrs": XDC file is prepended to the contents of constrs.xdc
        # 3. "post_constrs": XDC file is appended to the contents of constrs.xdc
        for xdc_file in self.target.content.xdc_files:
            if xdc_file.xdc_mode == 'read_xdc':
                for file in xdc_file.files:
                    self.writeln(f'read_xdc "{back2fwd(file)}"')

        if not self.target.cfg.custom_top:
            # write constraints to file
            constrs = CodeGenerator()

            # read in "pre_constr" XDC files
            for xdc_file in self.target.content.xdc_files:
                if xdc_file.xdc_mode == 'pre_constr':
                    for file in xdc_file.files:
                        constrs.writeln(f'# pre_constr: {file}')
                        constrs.read_from_file(file)
                        constrs.writeln()

            # generate constraints for external clk
            constrs.use_templ(TemplExtClk(target=self.target))
            # generate clock wizard IP core
            self.use_templ(TemplClkWiz(target=self.target))

            # Add IP cores necessary for control interface
            ip_core_templates = self.target.ctrl.add_ip_cores(scfg=scfg, ip_dir=self.target.ip_dir)
            for ip_core_template in ip_core_templates:
                self.use_templ(ip_core_template)

            ## Add constraints for additional generated emu_clks
            constrs.writeln('create_generated_clock -name emu_clk -source [get_pins clk_gen_i/clk_wiz_0_i/clk_out1] -divide_by 2 [get_pins gen_emu_clks_i/buf_emu_clk/I]')

            for k in range(scfg.num_gated_clks):
                constrs.writeln(f'create_generated_clock -name clk_other_{k} -source [get_pins clk_gen_i/clk_wiz_0_i/clk_out1] -divide_by 4 [get_pins gen_emu_clks_i/buf_{k}/I]')

            # Setup ILA for signal probing - only of at least one probe is defined
            if len(scfg.analog_probes + scfg.digital_probes + [scfg.time_probe]) != 0:
                self.use_templ(TemplILA(target=self.target, depth=self.target.prj_cfg.ila_depth))
    
            # Setup Debug Hub
            constrs.use_templ(TemplDbgHub(target=self.target))

            # Add false paths for Zynq control signals.  This is necessary in some cases
            # to provide timing violations, since the ARM core is running on a different
            # clock that the emulator circuitry.  This is a real problem, but is handled
            # in firmware with handshaking and very short delays.
            if self.target.cfg.fpga_sim_ctrl == FPGASimCtrl.UART_ZYNQ:
                constrs.writeln('set_false_path -through [get_pins sim_ctrl_gen_i/zynq_gpio_i/*]')

            # read in "post_constr" XDC files
            for xdc_file in self.target.content.xdc_files:
                if xdc_file.xdc_mode == 'post_constr':
                    for file in xdc_file.files:
                        constrs.writeln(f'# post_constr: {file}')
                        constrs.read_from_file(file)
                        constrs.writeln()

            # write master constraints to file and add to project
            master_constr_path = os.path.join(self.target.prj_cfg.build_root, 'constrs.xdc')
            constrs.write_to_file(master_constr_path)
            self.add_files([master_constr_path], fileset='constrs_1')

        # read user-provided IPs
        self.writeln('# Custom user-provided IP cores')
        for xci_file in self.target.content.xci_files:
            for file in xci_file.files:
                self.writeln(f'read_ip "{back2fwd(file)}"')

        # read user-provided TCL scripts
        self.writeln('# Custom user-provided TCL scripts')
        for tcl_file in self.target.content.tcl_files:
            for file in tcl_file.files:
                self.writeln(f'source "{back2fwd(file)}"')

        # upgrade IPs as necessary
        self.writeln('if {[get_ips] ne ""} {')
        self.writeln('    upgrade_ip [get_ips]')
        self.writeln('}')

        # generate all IPs
        self.writeln('generate_target all [get_ips]')

        # create additional Hardware for control interface
        if self.target.cfg.fpga_sim_ctrl == FPGASimCtrl.UART_ZYNQ:
            self.use_templ(TemplZynqGPIO(is_ultrascale=scfg.is_ultrascale))

        # launch the build and wait for it to finish
        num_cores = min(int(self.target.prj_cfg.vivado_config.num_cores), 8)
        self.writeln(f'launch_runs impl_1 -to_step write_bitstream -jobs {num_cores}')
        self.writeln('wait_on_run impl_1')

        # re-generate the LTX file
        # without this step, the ILA probes are sometimes split into individual bits
        impl_dir = os.path.join(
            project_root,
            f'{self.target.prj_cfg.vivado_config.project_name}.runs',
            'impl_1',
        )

        ltx_file_path = os.path.join(
            impl_dir, f"{self.target.cfg.top_module}.ltx"
        )

        self.writeln('open_run impl_1')
        self.writeln(f'write_debug_probes -force {{{back2fwd(ltx_file_path)}}}')

        # export the XSA if this is a more recent version of vivado
        if self.version_year >= 2020:
            xsa_file_path = os.path.join(impl_dir, f'{self.target.cfg.top_module}.xsa')
            xsa_file_path = back2fwd(xsa_file_path)
            self.writeln(f'write_hw_platform -fixed -include_bit -force -file {{{xsa_file_path}}}')

        #remove and restore drive substitutions
        if self.subst:
            self.writeln('exec subst ' + self.subst + ' /d')
            if self.old_subst:
                self.writeln('exec subst ' + self.subst + ' ' + self.old_subst)

        # run bitstream generation
        err_str = 'The design failed to meet the timing requirements.'
        try:
            self.run(filename=r"bitstream.tcl", stack=self.target.prj_cfg.cfg.vivado_stack, err_str=err_str)
        except:
            # remove and restore drive substitutions
            if os.name == 'nt':
                if self.subst:
                    try:
                        subprocess.call(f'subst {drive} /d', shell=True)
                    except:
                        print(f'WARNING: Removing mapped drive:{drive} did not work.')
                    if self.old_subst:
                        try:
                            subprocess.call(f'subst {drive} {self.old_subst}', shell=True)
                        except:
                            print(f'WARNING: Mapping of drive:{drive} to network path: {self.old_subst} did not work.')

            # then re-raise the original exception
            raise


    def run_FPGA(self, **kwargs):
        """
        Run the FPGA in non-interactive mode. This means FPGA will run for a specified
        duration; all specified signals will be captured and dumped to a file.

        See anasymod/templates/execute_FPGA_sim.py for keyword arguments.
        """

        self.use_templ(TemplEXECUTE_FPGA_SIM(target=self.target, **kwargs))
        self.run(filename='run_FPGA.tcl', interactive=False)

    def launch_FPGA(self, server_addr: str) -> CtrlApi:
        """
        Run the FPGA in interactive mode. This means FPGA will be programmed and
        control interfaces prepared. After that interactive communication with FPGA
        is possible.

        :param server_addr: Hardware server address for HW server launched by Vivado
        """

        self.target.ctrl_api._initialize()
        self.target.ctrl_api._setup_ctrl(server_addr=server_addr)

        # return ctrl object to user for further interactive commands
        return self.target.ctrl_api



from copy import deepcopy
from anasymod.templates.templ import JinjaTempl
from anasymod.generators.gen_api import SVAPI, ModuleInst
from anasymod.sim_ctrl.datatypes import DigitalSignal


class ModuleRegMapSimCtrl(JinjaTempl):
    def __init__(self, scfg):
        super().__init__(trim_blocks=True, lstrip_blocks=True)

        crtl_inputs = scfg.analog_ctrl_inputs + scfg.digital_ctrl_inputs
        ctrl_outputs = scfg.analog_ctrl_outputs + scfg.digital_ctrl_outputs

        #####################################################
        # Define module ios
        #####################################################
        self.module_ifc = SVAPI()
        module = ModuleInst(api=self.module_ifc, name="reg_map")

        # Add clock
        clk = DigitalSignal(name='clk', width=1, abspath=None)
        module.add_input(io_obj=clk)

        o_ctrl = DigitalSignal(name='o_ctrl', width=32, abspath=None)
        module.add_input(io_obj=o_ctrl)
        o_data = DigitalSignal(name='o_data', width=32, abspath=None)
        module.add_output(io_obj=o_data)
        i_ctrl = DigitalSignal(name='i_ctrl', width=32, abspath=None)
        module.add_input(io_obj=i_ctrl)
        i_data = DigitalSignal(name='i_data', width=32, abspath=None)
        module.add_input(io_obj=i_data)

        # Add I/Os to Design (probes and Conrol Parameters)
        module.add_outputs(io_objs=crtl_inputs, connections=crtl_inputs)
        module.add_inputs(io_objs=ctrl_outputs, connections=ctrl_outputs)

        module.generate_header()

        #####################################################
        # Initialize Default Values for ControlInfrastructure Parameters
        #####################################################

        self.init_ctrlios = SVAPI()

        for parameter in crtl_inputs:
            default_signal = deepcopy(parameter)
            default_signal.name = str(default_signal.name) + '_def'
            self.init_ctrlios.gen_signal(io_obj=default_signal)
            self.init_ctrlios.assign_to(io_obj=default_signal, exp=default_signal.init_value)

        #####################################################
        # Combo mux for Probes section
        #####################################################

        # instantiation of register map module
        self.probes_combomux_cases = SVAPI()

        self.probes_combomux_cases.indent(quantity=3)
        for probe in ctrl_outputs:
            if probe.o_addr is not None:
                self.probes_combomux_cases.writeln(f"'d{probe.o_addr}: o_data_reg = {probe.name};")

        #####################################################
        # Combo mux for Probes section
        #####################################################

        # instantiation of register map module
        self.params_regmap = SVAPI()

        for param in crtl_inputs:
            # create a reg signal
            reg_signal = deepcopy(param)
            reg_signal.name = f'{param.name}_reg'
            self.params_regmap.gen_signal(reg_signal)

            # assign to the reg signal
            self.params_regmap.writeln(f'assign {param.name} = {param.name}_reg;')

            # update the reg signal
            self.params_regmap.writeln(f'''\
always @(posedge clk) begin
    if (i_rst == 'b1) begin
        {param.name}_reg <= {param.name}_def; // use VIO defaults
    end else if ((i_valid == 1'b1) && (i_addr == 'd{param.i_addr})) begin
        {param.name}_reg <= i_data;
    end else begin
        {param.name}_reg <= {param.name}_reg;
    end
end''')

    TEMPLATE_TEXT = '''
`timescale 1ns/1ps

`default_nettype none

{{subst.module_ifc.text}}

    // break out signals in o_ctrl    
    logic [7:0] o_addr;
    assign o_addr = o_ctrl[7:0];
    
    // break out signals in i_ctrl
    logic i_rst;
    assign i_rst = i_ctrl[31];
    
    logic i_valid;
    assign i_valid = i_ctrl[30];
    
    logic [7:0] i_addr;
    assign i_addr = i_ctrl[7:0];
    
    // Initial values for parameters
    {{subst.init_ctrlios.text}}

	// combo mux for reading outputs from design
	logic [31:0] o_data_reg;
	assign o_data = o_data_reg;
	always @* begin
		case (o_addr)
{{subst.probes_combomux_cases.text}}
		    default: o_data_reg = 0;
		endcase
	end

	// register map for writing to the inputs of the design
	{{subst.params_regmap.text}}

endmodule
`default_nettype wire
'''


# general imports
import os
import os.path
import sys
import json
import shlex
import re
from multiprocessing.pool import ThreadPool
from math import ceil, log2
from collections import namedtuple
from argparse import ArgumentParser
from subprocess import Popen, PIPE, STDOUT
from glob import glob
from typing import Union, Pattern

# msdsl imports
from msdsl.generator.verilog import VerilogGenerator


def back2fwd(path: str):
    return path.replace('\\', '/')

def error_detected(text, err_str):
    # Returns True if the error pattern "err_str" is detected in "text"
    # "err_str" can be a one of several things:
    # 1. A single string.
    # 2. A list of strings.  The error is considered to be found if any of the
    #    strings appear in the given text.
    # 3. A regular expression Pattern (re.Pattern).  The given text is searched
    #    for any occurrence of this pattern.
    # from: https://github.com/leonardt/fault/blob/master/fault/subprocess_run.py
    if isinstance(err_str, str):
        return err_str in text
    elif isinstance(err_str, list):
        return any(elem in text for elem in err_str)
    elif isinstance(err_str, Pattern):
        return err_str.match(text) is not None
    else:
        raise Exception(f'Invalid err_str: {err_str}.')

def tee_output(fd, err_str=None):
    # prints lines from the given file descriptor while checking for errors
    # returns a flag indicating whether an error was detected
    # modified from: https://github.com/leonardt/fault/blob/master/fault/subprocess_run.py
    found_err = None
    for line in fd:
        # display line
        print(line, end='')
        # check line for errors
        if err_str is not None:
            if error_detected(text=line, err_str=err_str):
                found_err = line
    # Return flag indicating whether an error was found
    return found_err

def call(args, cwd=None, wait=True, err_str=None, return_error=False):
    # run a command and optionally check for error strings in the output
    # modified from: https://github.com/leonardt/fault/blob/master/fault/subprocess_run.py
    # print command string with proper escaping so that
    # a user can simply copy and paste the command to re-run it
    cmd_str = ' '.join(shlex.quote(arg) for arg in args)
    print(f"Checking return code of subprocess call: {cmd_str}")
    # run the command
    if wait:
        with Popen(args, cwd=cwd, stdout=PIPE, stderr=STDOUT, bufsize=1,
                   universal_newlines=True) as p:
            # print output while checking for errors
            found_err = tee_output(fd=p.stdout, err_str=err_str)

            # get return code and check result if desired
            returncode = p.wait()

            # check return code
            assert returncode == 0, f'Exited with non-zero code: {returncode}'

            # check for an error in the output text
            if found_err is not None:
                if return_error:
                    return found_err
                else:
                    raise OutputError(f'Found {err_str} in output of subprocess: {found_err}')
            else:
                return 0
    else:
        Popen(args=args, cwd=cwd, stdout=sys.stdout, stderr=sys.stdout)


class OutputError(Exception):
    def __init__(self, *args, **kwargs):
        pass

def next_pow_2(x):
    '''
    Return y such that 2**y >= x
    '''
    return 2**int(ceil(log2(x)))

def expand_searchpaths(paths: Union[list, str], rel_path_reference: str):
    abs_paths = []

    if isinstance(paths, str):
        paths = [paths]
    elif isinstance(paths, list):
        pass
    else:
        raise TypeError(f"Wrong format used for passing searchpaths, expecting list: {paths}")
    for file in paths:
        file = expand_path(path=file, rel_path_reference=rel_path_reference)
        abs_paths.append(file)

    return [file for p in abs_paths for file in glob(p)]

def expand_path(path, rel_path_reference):
    path = os.path.expandvars(str(path).strip('" '))
    if not os.path.isabs(path):
        if rel_path_reference is not None:
            path_suffix = path.replace('\\', '/').replace('/', os.sep).split(os.sep)
            try:
                path_suffix.remove('.')
            except:
                pass
            path = os.path.join(rel_path_reference, *(path_suffix))
        else:
            raise KeyError(f"No reference for expanding relative paths was provided for path: {path}")
    return path

########################
# parallel_scripts
# ref: https://stackoverflow.com/questions/26774781/python-multiple-subprocess-with-a-pool-queue-recover-output-as-soon-as-one-finis

def parallel_calls(calls, num=None):
    tp = ThreadPool(num)

    for arg_list in calls:
        tp.apply_async(call, (arg_list,))

    tp.close()
    tp.join()

########################
# file_len: modified from https://stackoverflow.com/questions/845058/how-to-get-line-count-cheaply-in-python

def file_len(fname):
    with open(fname, encoding='utf-8') as f:
        i = 0

        for i, l in enumerate(f, 1):
            pass

        return i


def vivado_search_key(dir_):
    '''
    Determine the year and version of a Vivado install directory.
    '''
    year, version = os.path.basename(dir_).split('.')
    year, version = int(year), int(version)

    return -year, -version

########################
# JSON to object: from https://stackoverflow.com/questions/6578986/how-to-convert-json-data-into-a-python-object

def _json_object_hook(d):
    return namedtuple('X', d.keys())(*d.values())
def json2obj(data):
    return json.loads(data, object_hook=_json_object_hook)
########################


# Argument parser for the examples
class ExampleControl:
    def __init__(self):
        # create the parser
        parser = ArgumentParser()

        # add custom arguments
        parser.add_argument('-o', '--output', type=str)
        parser.add_argument('--dt', type=float)

        # parser arguments
        args = parser.parse_args()

        # save arguments
        self.output = os.path.realpath(os.path.expanduser(args.output))
        self.dt = args.dt

    def write_model(self, model):
        # determine the filename
        filename = os.path.join(self.output, f'{model.module_name}.sv')
        print('Model will be written to: ' + filename)

        # write the model to a file
        model.compile_to_file(VerilogGenerator(), filename)


def main():
    print(next_pow_2(15))
    print(next_pow_2(16))
    print(next_pow_2(17))


if __name__ == '__main__':
    main()



from anasymod.util import back2fwd

# References:
# 1. https://www.xilinx.com/html_docs/xilinx2018_1/SDK_Doc/xsct/use_cases/xsdb_standalone_app_debug.html
# 2. https://github.com/Digilent/Arty-Z7-20-linux_bd/blob/master/sdk/.sdk/launch_scripts/xilinx_c-c%2B%2B_application_(system_debugger)/system_debugger_using_debug_video.elf_on_local.tcl
# 3. https://xilinx-wiki.atlassian.net/wiki/spaces/A/pages/18842240/Programming+QSPI+from+U-boot+ZC702
# 4. https://github.com/analogdevicesinc/no-OS/blob/master/scripts/xsdb.tcl
# 5. https://www.xilinx.com/html_docs/xilinx2018_1/SDK_Doc/xsct/use_cases/xsdb_debug_app_zynqmp.html

from pathlib import Path

class TemplXSCTProgram:
    def __init__(self, sdk_path, bit_path, hw_path, tcl_path, pcfg,
                 sw_name='sw', program_fpga=True, reset_system=True,
                 loadhw=True, init_cpu=True, download=True,
                 run=True, connect=True, is_ultrascale=False,
                 xsct_install_dir=None, no_rev_check=False,
                 server_addr=None):

        # save settings
        self.sdk_path = sdk_path
        self.bit_path = bit_path
        self.hw_path = hw_path
        self.tcl_path = tcl_path
        self.sw_name = sw_name
        self.is_ultrascale = is_ultrascale
        self.xsct_install_dir = xsct_install_dir
        self.no_rev_check = no_rev_check
        self.pcfg  = pcfg

        # initialize text
        self.text = ''

        ##################
        # apply commands #
        ##################

        # connect
        if connect:
            self.connect(server_addr=server_addr)

        # load utility functions for Zynq MP
        if is_ultrascale:
            self.puts('Load utility functions for Zynq MP.')
            zynqmp_utils_tcl = (Path(self.xsct_install_dir) / 'scripts' /
                                'vitis' / 'util' / 'zynqmp_utils.tcl')
            self.line(f'source "{zynqmp_utils_tcl.as_posix()}"')
            self.line()

        # reset_system
        if reset_system:
            self.reset_system()

        # program_fpga
        if program_fpga:
            self.program_fpga()

        # loadhw
        if loadhw:
            self.loadhw(hw_path)

        # disable access protection for dow, mrd, and mwr commands
        if is_ultrascale:
            self.puts('Disable access protection for dow, mrd, and mwr commands.')
            self.line('configparams force-mem-access 1')
            self.line()

        # init_cpu
        if init_cpu:
            self.init_cpu()

        # download
        if download:
            self.download()

        # enable access protection for dow, mrd, and mwr commands
        if is_ultrascale:
            self.puts('Enable access protection for dow, mrd, and mwr commands.')
            self.line('configparams force-mem-access 0')
            self.line()

        # run
        if run:
            self.run()

        # print message indicating that program has been loaded
        self.puts("Program started.")

    def line(self, s='', nl='\n'):
        self.text += f'{s}{nl}'

    def puts(self, s):
        self.line(f'puts "{s}"')

    def set_target(self, pattern):
        self.line(f'targets -set -filter {{name =~ "{pattern}"}}')

    def connect(self, server_addr):
        self.puts('Connecting to the HW server...')
        if server_addr is not None:
            self.line(f'connect -url {server_addr}')
        else:
            self.line('connect')
        self.line()

    def reset_system(self):
        self.puts('Resetting the system...')
        self.set_target('APU*')
        if self.is_ultrascale:
            self.line('rst -system')
            self.line('after 3000')
        else:
            self.line('stop')
            self.line('rst')
        self.line()

    def program_fpga(self):
        self.puts('Programming the FPGA...')
        if self.is_ultrascale:
            self.set_target('PSU')
        else:
            self.set_target('xc7z*')
        cmd = []
        cmd += ['fpga']
        cmd += ['-f', f'"{self.bit_path.as_posix()}"']
        if self.no_rev_check:
            cmd += ['-no-revision-check']
        self.line(' '.join(cmd))
        self.line()

    def loadhw(self, hw_path):
        if self.pcfg.vivado_config.version_year < 2020:
            # For Vivado 2018.2, it is necessary to use the .hdf rather than the .sysdef file.
            hw_path = hw_path.with_suffix('.hdf')

        self.puts('Setting up the debugger...')
        self.set_target('APU*')
        cmd = []
        cmd += ['loadhw']
        cmd += ['-hw', f'"{hw_path.as_posix()}"']
        if self.is_ultrascale:
            cmd += ['-mem-ranges [list {0x80000000 0xbfffffff} '
                                      '{0x400000000 0x5ffffffff} '
                                      '{0x1000000000 0x7fffffffff}]']
            cmd += ['-regs']
        self.line(' '.join(cmd))
        self.line()

    def init_cpu(self):
        self.puts('Initializing the processor...')
        if self.is_ultrascale:
            # set the boot mode
            self.set_target('APU*')
            self.line('set mode [expr [mrd -value 0xFF5E0200] & 0xf]')
            # reset the processor
            self.set_target('*A53*#0')
            self.line('rst -processor')
            # download the FSBL ELF is located
            fsbl_path = (Path(self.sdk_path) / 'top' / 'export' / 'top' /
                         'sw' / 'top' / 'boot' / 'fsbl.elf')
            self.line(f'dow "{fsbl_path.as_posix()}"')
            # wait for FSBL to finish running
            self.line('set bp_0_6_fsbl_bp [bpadd -addr &XFsbl_Exit]')
            self.line('con -block -timeout 60')
            self.line('bpremove $bp_0_6_fsbl_bp')
        else:
            self.set_target('APU*')
            self.line(f'source "{self.tcl_path.as_posix()}"')
            self.line('ps7_init')
            self.line('ps7_post_config')
        self.line()

    def download(self):
        elf_path = str(self.sdk_path / self.sw_name / 'Debug' / self.sw_name) + '.elf'
        self.puts('Downloading the program...')
        if self.is_ultrascale:
            self.set_target('*A53*#0')
            self.line('rst -processor')
        else:
            self.set_target('*Cortex-A9 MPCore #0*')
        self.line(f'dow "{back2fwd(elf_path)}"')
        self.line()

    def run(self):
        self.puts('Starting the program...')
        self.line('con')
        self.line()




__all__ = ["ParseVCD"]

class ParseVCD:
    def __init__(self, vcd_root):
        self.vcd_root = vcd_root
        self.cycle_value = 'cv'

    def parse_vcd(self, sig_names=0, stdout=0, sigs=[], update_data=False):
        #Initialization
        usigs = {}
        data = {}
        hierarchy = []
        cycle_cnt = 0
        
        for i in sigs:
            usigs[i] = 1

        if len(usigs):
            all_sigs = 0
        else:
            all_sigs = 1

        with open(self.vcd_root, 'r') as file:
            while True:
                line = file.readline()
                if line == '':  # End-of-File
                    self.update_data(cycle_cnt=cycle_cnt, data=data)
                    break
                line = line.strip()

                if line == '':
                    continue

                if line[0] in ('r', 'R', 'b', 'B'):
                    (value, code) = line[1:].split()
                    try:
                        #convert all x and z to 0
                        if line[0] in ('b', 'B'):
                            value = value.replace('x', '0').replace('z', '0')
                        else:
                            value = float(value)
                    except:
                        # This is a quick fix that sets X,Z, ... to 0
                        value = float(0)
                    if (code in data):
                        if (stdout):
                            print(cycle_cnt, value)
                        else:
                            if self.cycle_value not in data[code]:
                                data[code][self.cycle_value] = []
                            data[code][self.cycle_value].append((cycle_cnt, value))

                elif line[0] in ('x', 'X', 'z', 'Z', '0', '1'):
                    value = line[0]
                    code = line[1:]
                    if (code in data):
                        if (stdout):
                            print(cycle_cnt, value)
                        else:
                            if self.cycle_value not in data[code]:
                                data[code][self.cycle_value] = []
                            data[code][self.cycle_value].append((cycle_cnt, value))

                elif line[0] == '#':
                    if update_data: self.update_data(cycle_cnt=cycle_cnt, data=data)
                    cycle_cnt = int(line[1:])

                elif "$enddefinitions" in line:
                    num_sigs = len(data)
                    if (num_sigs == 0):
                        if (all_sigs):
                            raise Exception(f"No signals were found reading VCD file {self.vcd_root}")
                        else:
                            raise Exception(f"No matching signals were found reading VCD file: {self.vcd_root}")
                    if ((num_sigs > 1) and stdout):
                        raise Exception("Too many signals provided to VCD parser!")
                    if sig_names:
                        break

                elif "$scope" in line:
                    hierarchy.append(line.split()[2])

                elif "$upscope" in line:
                    hierarchy.pop()

                elif "$var" in line:
                    ls = line.split()
                    type = ls[1]
                    size = ls[2]
                    code = ls[3]

                    name = ls[4]
                    path = '.'.join(hierarchy)
                    full_name = path + '.' + name
                    if (full_name in usigs) or all_sigs:
                        if code not in data:
                            data[code] = {}
                        if 'nets' not in data[code]:
                            data[code]['nets'] = []
                        var_struct = {
                            'type': type,
                            'name': name,
                            'size': size,
                            'hier': path,
                        }
                        if var_struct not in data[code]['nets']:
                            data[code]['nets'].append(var_struct)

        file.close()

        return data
    
    def update_data(self, cycle_cnt, data):
        if cycle_cnt != 0:
            cycle_cnt_old = cycle_cnt
            for d in data.keys():
                length = len(data[d][self.cycle_value]) - 1
                if data[d][self.cycle_value][length][0] != cycle_cnt_old:
                    data[d][self.cycle_value].append((cycle_cnt_old, data[d][self.cycle_value][length][1]))

    def list_sigs(self):

        vcd = self.parse_vcd(sig_names=1)

        sigs = []
        for k in vcd.keys():
            v = vcd[k]
            nets = v['nets']
            sigs.extend(n['hier'] + '.' + n['name'] for n in nets)

        return sigs



import os
import pytest

def pytest_addoption(parser):
    parser.addoption("--rtenv", action="store", default="nocplusplus",
        help="Select for which Run Time Environment test should be executed.")
    parser.addoption("--classification", action="store", default="basic",
                     help="Select what classification of test should be executed.")
    parser.addoption("--target", action="store", default='emulate_vivado',
                     help="Select for which target the test should be executed.")

"""
@pytest.fixture( scope="session")
def setup_environment():
    return "helllo"

def pytest_namespace():
    return {'basic' : pytest.mark.usefixtures("basic")}


@pytest.fixture(scope="session")
def basic():
    print("yippi basic", pytest.config.getoption("--classification"))
    print(TestClassification.basic not in pytest.config.getoption("--classification"))
    basic = pytest.mark.skipif(
        TestClassification.basic not in pytest.config.getoption("--classification"),
        reason="need basic as argumentfor --classification",
    )
    return basic

@pytest.fixture
def weekend():
    print("weekend")
    return pytest.mark.skipif(
        TestClassification.weekend not in pytest.config.getoption("--classification")
    )
"""



class Signal():
    def __init__(self, abs_path: str, delimiter='.'):
        """
        Class for storing signal information, that will be later used to generate HDL code related to signals.
        :param abs_path:    Absolute path to the defined signal, this can also only be the name itself, in case it will
                            not be used for a connection via abs path later
        :param delimiter:   Delimiter to separate the signal's abs path
        """
        self.delimiter = delimiter
        # Separate the signal's abs path
        signal_hier = abs_path.split(self.delimiter)

        # Check if an abs path was actually provided or only the signal name
        if len(signal_hier) == 1:
            # Only signal name was given
            self.name = ''.join(signal_hier)
            self.abs_path = None
        elif len(signal_hier) > 1:
            # Abs path to signal was provided
            self.name = signal_hier[-1]
            self.abs_path = abs_path


from anasymod.templates.templ import JinjaTempl
from anasymod.config import EmuConfig
from anasymod.generators.gen_api import SVAPI, ModuleInst
from anasymod.structures.structure_config import StructureConfig
from anasymod.sim_ctrl.datatypes import DigitalSignal

class ModuleTracePort(JinjaTempl):
    def __init__(self, scfg: StructureConfig):
        super().__init__(trim_blocks=True, lstrip_blocks=True)
        probes = scfg.digital_probes + scfg.analog_probes + [scfg.time_probe] + [scfg.dec_cmp]

        #####################################################
        # Define module ios
        #####################################################

        self.module_ifc = SVAPI()
        module = ModuleInst(api=self.module_ifc, name="trace_port_gen")
        # Add probe signals
        module.add_inputs(probes)

        # Add master clk
        module.add_input(scfg.emu_clk)

        module.generate_header()

        #####################################################
        # CPU sim control section - add dump statements
        #####################################################

        self.probe_dumps = SVAPI()
        self.probe_dumps.indent()
        self.probe_dumps.writeln(f'initial begin')
        self.probe_dumps.indent()
        self.probe_dumps.writeln('#0;')
        for probe in probes:
            self.probe_dumps.writeln(f'$dumpvars(0, {probe.name});')
        self.probe_dumps.dedent()
        self.probe_dumps.writeln(f'end')

        #####################################################
        # FPGA sim control section - Instantiate ila core
        #####################################################

        # Instantiate ila core
        self.ila_wiz_inst = SVAPI()
        ila_wiz = ModuleInst(api=self.ila_wiz_inst, name="ila_0")
        for k, signal in enumerate(probes): # Add probe signals
            ila_wiz.add_input(DigitalSignal(name=f'probe{k}', abspath=None, width=signal.width), connection=signal)
        ila_wiz.add_input(DigitalSignal(name='clk', abspath=None, width=1), connection=scfg.emu_clk) # Add master clk
        ila_wiz.generate_instantiation()


    TEMPLATE_TEXT = '''
`timescale 1ns/1ps

`default_nettype none
{{subst.module_ifc.text}}

`ifdef SIMULATION_MODE_MSDSL
{{subst.probe_dumps.text}}
`else
	// ILA instantiation
{{subst.ila_wiz_inst.text}}

`endif // `ifdef SIMULATION_MODE_MSDSL

endmodule
`default_nettype wire
'''

def main():
    print(ModuleTracePort(scfg=StructureConfig(prj_cfg=EmuConfig(root='test', cfg_file=''))).render())

if __name__ == "__main__":
    main()


import os, yaml
from anasymod.sources import (Sources, VerilogSource, VerilogHeader, VHDLSource,
                              SubConfig, XCIFile, XDCFile, MEMFile, BDFile,
                              FunctionalModel, IPRepo, EDIFFile, FirmwareFile,
                              TCLFile, IncludeDir)
from anasymod.defines import Define
from anasymod.util import expand_searchpaths

class Filesets():
    def __init__(self, root, default_filesets, root_func_models):
        self._master_cfg_path = os.path.join(root, 'source.yaml')
        self.default_filesets = default_filesets
        self.root_func_models = root_func_models

        # Store all config file paths in list for packing purposes
        self._config_paths = [self._master_cfg_path]

        self._verilog_sources = []
        """:type : List[VerilogSource]"""

        self._verilog_headers = []
        """:type : List[VerilogHeader]"""

        self._vhdl_sources = []
        """:type : List[VHDLSource]"""

        self._include_dirs = []
        """:type : List[IncludeDir]"""

        self._edif_files = []
        """:type : List[EDIFFile]"""

        self._firmware_files = []
        """:type : List[FirmwareFile]"""

        self._xci_files = []
        """:type : List[XCIFile]"""

        self._tcl_files = []
        """:type : List[TCLFile]"""

        self._xdc_files = []
        """:type : List[XDCFile]"""

        self._defines = []
        """:type : List[Define]"""

        self._mem_files = []
        """:type : List[MEMFile]"""

        self._bd_files = []
        """:type : List[BDFile]"""

        self._ip_repos = []
        """:type : List[IPRepo]"""

        self._functional_models = []
        """:type : List[FunctionalModel]"""

        # init fileset_dict
        self.fileset_dict = {}
        if not default_filesets:
            for fileset in default_filesets:
                self.fileset_dict[fileset] = {}

        self._sub_config_paths = []
        """:type : List[SubConfig]"""

    def read_filesets(self, validate_paths=False):
        if os.path.isfile(self._master_cfg_path):
            try:
                mcfg = yaml.safe_load(open(self._master_cfg_path, "r"))
            except yaml.YAMLError as exc:
                raise Exception(exc)

            # Read source paths from master config
            self._parseconfig(cfg=mcfg, cfg_path=self._master_cfg_path)

            # Read source paths from sub configs
            while (bool(self._sub_config_paths)):
                for config in self._sub_config_paths:
                    self._sub_config_paths.remove(config)
                    files = expand_searchpaths(config.files, rel_path_reference=os.path.dirname(config.config_path))
                    for file in files:

                        if os.path.isfile(file):
                            try:
                                cfg = yaml.safe_load(open(file, "r"))
                            except yaml.YAMLError as exc:
                                raise Exception(exc)
                            self._parseconfig(cfg=cfg, cfg_path=file)
                            self._config_paths.append(config)  #Store processed config path
                        else:
                            print(f"WARNING: provided path:'{file}' does not exist, skipping config file")

            # Check if fileset paths exist
            if validate_paths:
                for key in self.fileset_dict.keys():
                    for path in self.fileset_dict[key]:
                        if not os.path.exists(path):
                            raise ValueError(f"Provided path:{path} of fileset:{key} does not exist")
        else:
            print(f"No config file existing, skipping to read source files.")

    def _parseconfig(self, cfg: dict, cfg_path: str):
        """
        Store all entries from source.yaml to the according fileset or added to the list of additional config paths,
        which will be read and stored in the next iteration.
        :param cfg: source.yaml stored as dict
        :param cfg_path: path to source.yaml file
        """

        if 'verilog_sources' in cfg.keys(): # Add verilog sources to filesets
            print(f'Verilog Sources: {[key for key in cfg["verilog_sources"].keys()]}')
            for verilog_source in cfg['verilog_sources'].keys():
                self._verilog_sources.append(VerilogSource(files=cfg['verilog_sources'][verilog_source]['files'],
                                                           fileset=cfg['verilog_sources'][verilog_source]['fileset'] if 'fileset' in cfg['verilog_sources'][verilog_source].keys() else 'default',
                                                           config_path=cfg_path,
                                                           version=cfg['verilog_sources'][verilog_source]['version'] if 'version' in cfg['verilog_sources'][verilog_source].keys() else None,
                                                           name=verilog_source))
        if 'verilog_headers' in cfg.keys(): # Add verilog headers to filesets
            print(f'Verilog Headers: {[key for key in cfg["verilog_headers"].keys()]}')
            for verilog_header in cfg['verilog_headers'].keys():
                self._verilog_headers.append(VerilogHeader(files=cfg['verilog_headers'][verilog_header]['files'],
                                                           fileset=cfg['verilog_headers'][verilog_header]['fileset'] if 'fileset' in cfg['verilog_headers'][verilog_header].keys() else 'default',
                                                           config_path=cfg_path,
                                                           name=verilog_header))
        if 'vhdl_sources' in cfg.keys(): # Add VHDL sources to filesets
            print(f'VHDL Sources: {[key for key in cfg["vhdl_sources"].keys()]}')
            for vhdl_source in cfg['vhdl_sources'].keys():
                self._vhdl_sources.append(VHDLSource(files=cfg['vhdl_sources'][vhdl_source]['files'],
                                                     fileset=cfg['vhdl_sources'][vhdl_source]['fileset'] if 'fileset' in cfg['vhdl_sources'][vhdl_source].keys() else 'default',
                                                     config_path=cfg_path,
                                                     library=cfg['vhdl_sources'][vhdl_source]['library'] if 'library' in cfg['vhdl_sources'][vhdl_source].keys() else None,
                                                     version = cfg['vhdl_sources'][vhdl_source]['version'] if 'version' in cfg['vhdl_sources'][vhdl_source].keys() else None,
                                                     name=vhdl_source))
        if 'include_dirs' in cfg.keys(): # Add include dirs to filesets
            print(f'Include Dirs: {[key for key in cfg["include_dirs"].keys()]}')
            for include_dir in cfg['include_dirs'].keys():
                self._include_dirs.append(IncludeDir(files=cfg['include_dirs'][include_dir]['files'],
                                                     fileset=cfg['include_dirs'][include_dir].get('fileset', 'default'),
                                                     config_path=cfg_path,
                                                     name=include_dir))
        if 'edif_files' in cfg.keys(): # Add EDIF files to filesets
            print(f'EDIF Files: {[key for key in cfg["edif_files"].keys()]}')
            for edif_file in cfg['edif_files'].keys():
                self._edif_files.append(EDIFFile(files=cfg['edif_files'][edif_file]['files'],
                                                fileset=cfg['edif_files'][edif_file]['fileset'] if 'fileset' in cfg['edif_files'][edif_file].keys() else 'default',
                                                config_path=cfg_path,
                                                name=edif_file))
        if 'firmware_files' in cfg.keys(): # Add firmware files to filesets
            print(f'Firmware Files: {[key for key in cfg["firmware_files"].keys()]}')
            for firmware_file in cfg['firmware_files'].keys():
                self._firmware_files.append(FirmwareFile(files=cfg['firmware_files'][firmware_file]['files'],
                                                fileset=cfg['firmware_files'][firmware_file]['fileset'] if 'fileset' in cfg['firmware_files'][firmware_file].keys() else 'default',
                                                config_path=cfg_path,
                                                name=firmware_file))
        if 'xci_files' in cfg.keys(): # Add xci files to filesets
            print(f'XCI Files: {[key for key in cfg["xci_files"].keys()]}')
            for xci_file in cfg['xci_files'].keys():
                self._xci_files.append(XCIFile(files=cfg['xci_files'][xci_file]['files'],
                                               fileset=cfg['xci_files'][xci_file]['fileset'] if 'fileset' in cfg['xci_files'][xci_file].keys() else 'default',
                                               config_path=cfg_path,
                                               name=xci_file))
        if 'tcl_files' in cfg.keys(): # Add TCL files to filesets
            print(f'TCL Files: {[key for key in cfg["tcl_files"].keys()]}')
            for tcl_file in cfg['tcl_files'].keys():
                self._tcl_files.append(TCLFile(files=cfg['tcl_files'][tcl_file]['files'],
                                               fileset=cfg['tcl_files'][tcl_file]['fileset'] if 'fileset' in cfg['tcl_files'][tcl_file].keys() else 'default',
                                               config_path=cfg_path,
                                               name=tcl_file))
        if 'xdc_files' in cfg.keys(): # Add constraint files to filesets
            print(f'XDC Files: {[key for key in cfg["xdc_files"].keys()]}')
            for xdc_file in cfg['xdc_files'].keys():
                self._xdc_files.append(XDCFile(files=cfg['xdc_files'][xdc_file]['files'],
                                               fileset=cfg['xdc_files'][xdc_file]['fileset'] if 'fileset' in cfg['xdc_files'][xdc_file].keys() else 'default',
                                               config_path=cfg_path,
                                               name=xdc_file,
                                               xdc_mode=cfg['xdc_files'][xdc_file].get('xdc_mode', None)))
        if 'mem_files' in cfg.keys():  # Add mem files to filesets
            print(f'Mem Files: {[key for key in cfg["mem_files"].keys()]}')
            for mem_file in cfg['mem_files'].keys():
                self._mem_files.append(MEMFile(files=cfg['mem_files'][mem_file]['files'],
                                               fileset=cfg['mem_files'][mem_file]['fileset'] if 'fileset' in cfg['mem_files'][mem_file].keys() else 'default',
                                               config_path=cfg_path,
                                               name=mem_file))
        if 'bd_files' in cfg.keys():  # Add block diagram files to filesets
            print(f'Block Diagram Files: {[key for key in cfg["bd_files"].keys()]}')
            for bd_file in cfg['bd_files'].keys():
                self._bd_files.append(BDFile(files=cfg['bd_files'][bd_file]['files'],
                                             fileset=cfg['bd_files'][bd_file]['fileset'] if 'fileset' in cfg['bd_files'][bd_file].keys() else 'default',
                                             config_path=cfg_path,
                                             name=bd_file))
        if 'ip_repos' in cfg.keys():  # Add IP repositories to filesets
            print(f'IP Repos: {[key for key in cfg["ip_repos"].keys()]}')
            for ip_repo in cfg['ip_repos'].keys():
                self._ip_repos.append(IPRepo(files=cfg['ip_repos'][ip_repo]['files'],
                                             fileset=cfg['ip_repos'][ip_repo]['fileset'] if 'fileset' in cfg['ip_repos'][ip_repo].keys() else 'default',
                                             config_path=cfg_path,
                                             name=ip_repo))
        if 'functional_models' in cfg.keys():  # Add functional model to filesets
            print(f'Functional Models: {[key for key in cfg["functional_models"].keys()]}')
            for functional_model in cfg['functional_models'].keys():
                func_model = FunctionalModel(files=cfg['functional_models'][functional_model]['files'],
                                             fileset=cfg['functional_models'][functional_model]['fileset'] if 'fileset' in cfg['functional_models'][functional_model].keys() else 'default',
                                             config_path=cfg_path,
                                             name=functional_model)
                func_model.set_gen_files_path(hdl_dir_root=self.root_func_models)
                self._functional_models.append(func_model)
        if 'sub_configs' in cfg.keys():  # Add sub config files to filesets
            for sub_config in cfg['sub_configs'].keys():
                self._sub_config_paths.append(SubConfig(files=cfg['sub_configs'][sub_config]['files'], config_path=cfg_path, name=sub_config))
        if 'defines' in cfg.keys():  # Add defines to filesets
            print(f'Defines: {[key for key in cfg["defines"].keys()]}')
            for define in cfg['defines'].keys():
                self._defines.append(Define(name=cfg['defines'][define]['name'],
                                            value=cfg['defines'][define]['value'] if 'value' in cfg['defines'][define].keys() else None,
                                            fileset=cfg['defines'][define]['fileset'] if 'fileset' in cfg['defines'][define].keys() else 'default'))

    def _expand_source_paths(self, sources: list):
        for source in sources:
            source.expand_paths()
            # Expand paths for generated files as well, if possible
            try:
                expand_gen_files_path = getattr(source, 'expand_gen_files_path')
            except:
                expand_gen_files_path = None

            if expand_gen_files_path:
                source.expand_gen_files_path()
        return sources

    def populate_fileset_dict(self):
        """
        Creates fileset dictionary according to filesets that were provided reading in source and define objects.
        Previously created filesets will be overwritten.
        """

        # Initialize all filesets
        self.fileset_dict = {}
        if not self.default_filesets:
            for fileset in self.default_filesets:
                self.fileset_dict[fileset] = {}

        # Read in verilog source objects to fileset dict
        self._add_to_fileset_dict(name='verilog_sources', container=self._expand_source_paths(self._verilog_sources))

        # Read in verilog header objects to fileset dict
        self._add_to_fileset_dict(name='verilog_headers', container=self._expand_source_paths(self._verilog_headers))

        # Read in vhdlsource objects to fileset dict
        self._add_to_fileset_dict(name='vhdl_sources', container=self._expand_source_paths(self._vhdl_sources))

        # Read in include_dirs objects to fileset dict
        self._add_to_fileset_dict(name='include_dirs', container=self._expand_source_paths(self._include_dirs))

        # Read in edif_files objects to fileset dict
        self._add_to_fileset_dict(name='edif_files', container=self._expand_source_paths(self._edif_files))

        # Read in firmware_files objects to fileset dict
        self._add_to_fileset_dict(name='firmware_files', container=self._expand_source_paths(self._firmware_files))

        # Read in xcifile objects to fileset dict
        self._add_to_fileset_dict(name='xci_files', container=self._expand_source_paths(self._xci_files))

        # Read in tclfile objects to fileset dict
        self._add_to_fileset_dict(name='tcl_files', container=self._expand_source_paths(self._tcl_files))

        # Read in xdcfile objects to fileset dict
        self._add_to_fileset_dict(name='xdc_files', container=self._expand_source_paths(self._xdc_files))

        # Read in define objects to fileset dict
        self._add_to_fileset_dict(name='defines', container=self._defines)

        # Read in mem_files objects to fileset dict
        self._add_to_fileset_dict(name='mem_files', container=self._expand_source_paths(self._mem_files))

        # Read in bd_files objects to fileset dict
        self._add_to_fileset_dict(name='bd_files', container=self._expand_source_paths(self._bd_files))

        # Read in ip_repos objects to fileset dict
        self._add_to_fileset_dict(name='ip_repos', container=self._expand_source_paths(self._ip_repos))

        # Read in functional model objects to fileset dict
        self._add_to_fileset_dict(name='functional_models', container=self._expand_source_paths(self._functional_models))

    def _add_to_fileset_dict(self, name, container):
        """
        Adds a specified attribute to the fileset_dict, e.g. add the verilog sources or defines.
        """
        for item in container:
            if item.fileset in self.fileset_dict.keys():
                if name in self.fileset_dict[item.fileset]:
                    self.fileset_dict[item.fileset][name].append(item)
                else:
                    self.fileset_dict[item.fileset][name] = [item]
            else:
                print(f"Custom fileset was added:{item.fileset}")
                self.fileset_dict[item.fileset] = {}
                self.fileset_dict[item.fileset][name] = [item]


    def add_source(self, source: Sources):
        if isinstance(source, VerilogSource):
            self._verilog_sources.append(source)
        if isinstance(source, VerilogHeader):
            self._verilog_headers.append(source)
        if isinstance(source, VHDLSource):
            self._vhdl_sources.append(source)

    def add_define(self, define: Define):
        self._defines.append(define)

    def add_include_dir(self, include_dir: IncludeDir):
        self._include_dirs.append(include_dir)

    def add_edif_file(self, edif_file: EDIFFile):
        self._edif_files.append(edif_file)

    def add_firmware_file(self, firmware_file: FirmwareFile):
        self._firmware_files.append(firmware_file)

    def add_xci_file(self, xci_file: XCIFile):
        self._xci_files.append(xci_file)

    def add_tcl_file(self, tcl_file: TCLFile):
        self._tcl_files.append(tcl_file)

    def add_xdc_file(self, xdc_file: XDCFile):
        self._xdc_files.append(xdc_file)

    def add_mem_file(self, mem_file: MEMFile):
        self._mem_files.append(mem_file)

    def add_bd_file(self, bd_file: BDFile):
        self._bd_files.append(bd_file)

    def add_ip_repo(self, ip_repo: IPRepo):
        self._ip_repos.append(ip_repo)

    def add_functional_model(self, functional_model: FunctionalModel):
        self._functional_models.append(functional_model)

def main():
    fileset = Filesets(root=r"C:\Inicio_dev\anasymod\tests\filter")
    fileset.read_filesets()
    print(fileset.fileset_dict)

if __name__ == '__main__':
    main()


from math import ceil, log2


class _Signal():
    """
    Container for signals that can be used to describe connectivity within the design.
    :param abs_path:    Absolute path to the defined signal, this can also only be the
                        name itself, in case it will not be used for a connection via
                        an absolute path later
    :param delimiter:   Delimiter to separate the signal's abs path
    :param name:        Port name
    """

    def __init__(self, abspath, name, delimiter='.'):
        self.delimiter = delimiter
        self.name = name
        self.abs_path = abspath

    @classmethod
    def from_dict(cls, name, dict):
        return cls(name=name, **dict)

# Digital Signal Classes


class DigitalSignal(_Signal):
    """
    Container for digital signals that can be used to describe connectivity within the design.
    Same instantiation arguments as _Signal, plus:
    :param width:       Number of bits
    _param signed:      If true, signal is of type signed
    """

    def __init__(self, *args, width=1, signed=False, **kwargs):
        # call the super constructor
        super().__init__(*args, **kwargs)
        # save settings
        self.width = width
        self.signed = signed


class DigitalCtrlInput(DigitalSignal):
    """
    Container for a digital control input to the simulation.
    Same instantiation arguments as DigitalSignal, plus:
    :param i_addr:      Address, which the control interface uses to access the corresponding entry in reg map
    :param init_value:  Initial value that will be transmitted via this input
    """

    def __init__(self, *args, init_value=0, i_addr=None, **kwargs):
        # call the super constructor
        super().__init__(*args, **kwargs)
        # save settings
        self.i_addr = i_addr
        self.init_value = init_value


class DigitalCtrlOutput(DigitalSignal):
    """
    Container for a digital control output to the simulation.
    Same instantiation arguments as DigitalSignal, plus:
    :param o_addr:      Address, which the control interface uses to access the corresponding entry in reg map
    """

    def __init__(self, *args, o_addr=None, **kwargs):
        # call the super constructor
        super().__init__(*args, **kwargs)
        # save settings
        self.o_addr = o_addr


# Analog Signal Classes


class AnalogSignal(_Signal):
    """
    Container for analog signals that can be used to describe connectivity within the design.
    Same instantiation arguments as _Signal, plus
    :param range:       Range for analog fixed-point datatype
    :param width:       Width for analog fixed-point datatype
    :param exponent:    Exponent for analog fixed-point datatype
    """

    def __init__(self, *args, range=None, width=25, exponent=None, **kwargs):
        # call the super constructor
        super().__init__(*args, **kwargs)

        # Set defaults for fixed-point formatting in case only two of the
        # three arguments have been specified.  Note that in order to cause
        # the width to be inferred, a user must explicity set width=None,
        # since the default value is "25"
        if (width is None) and ((range is not None) and (exponent is not None)):
            width = self.calc_width(range=range, exponent=exponent)
        elif (exponent is None) and ((range is not None) and (width is not None)):
            exponent = self.calc_exponent(range=range, width=width)
        elif (range is None) and ((width is not None) and (exponent is not None)):
            range = self.calc_range(width=width, exponent=exponent)

        # make sure the width and exponent are defined
        # (range is not strictly needed to interpret fixed-point types)
        assert width is not None, 'Width is not defined.'
        assert exponent is not None, 'Exponent is not defined.'
        assert range is not None, 'Range is not defined.'

        # save settings
        self.width = width
        self.exponent = exponent
        self.range = range

    @staticmethod
    def calc_exponent(range, width):
        if range == 0:
            return 0
        else:
            return int(ceil(log2(range/(2**(width-1)-1))))

    @staticmethod
    def calc_width(range, exponent):
        if range == 0:
            return 1
        else:
            return int(ceil(1+log2((range/(2**exponent))+1)))

    @staticmethod
    def calc_range(width, exponent):
        # should be defined such that the following expressions are true
        # e == calc_exponent(calc_range(w, e), w)
        # w == calc_width(calc_range(w, e), e)
        return ((2**(width-1))-1)*(2**exponent)


    def float_to_fixed(self, val):
        return int(round(val*(2**(-self.exponent))))

    def fixed_to_float(self, val):
        return float(val*(2**(self.exponent)))

class AnalogCtrlInput(AnalogSignal):
    """
    Container for an analog control input to the simulation.
    Same instantiation arguments as AnalogSignal, plus
    :param i_addr:      Address, which the control interface uses to access the corresponding entry in reg map
    :param init_value:  Initial value that will be transmitted via this input
    """

    def __init__(self, *args, init_value=0.0, i_addr=None, **kwargs):
        super().__init__(*args, **kwargs)
        self.i_addr = i_addr
        self.init_value = init_value


class AnalogCtrlOutput(AnalogSignal):
    """
    Container for an analog control output to the simulation.
    Same instantiation arguments as AnalogSignal, plus
    :param o_addr:      Address, which the control interface uses to access the corresponding entry in reg map
    """

    def __init__(self, *args, o_addr=None, **kwargs):
        super().__init__(*args, **kwargs)
        self.o_addr = o_addr


class AnalogProbe(AnalogSignal):
    """
    Container for an analog probe signals.
    Same instantiation arguments as AnalogSignal.
    """

    pass


class ProbeSignal():
    """
    Temporary container to store all the information necessary for generating ILA probe signals.
    """

    def __init__(self, name, abspath, width, exponent, type):
        self.name = name
        self.abspath = abspath
        self.width = width
        self.exponent = exponent
        self.type = type


from anasymod.templates.generic_ip import TemplGenericIp
from anasymod.util import next_pow_2
from anasymod.targets import FPGATarget

class TemplILA(TemplGenericIp):
    def __init__(self, target: FPGATarget, depth=4096):
        # set defaults
        scfg = target.str_cfg

        # Sanity checking for ILA depth
        assert next_pow_2(depth) == depth, 'The ILA depth must be a power of 2.'
        assert depth >= 1024, 'The ILA depth must be at least 1024.'

        self.inst_name = target.prj_cfg.vivado_config.ila_inst_name

        self.width_ila_clk = '1'
        self.conn_ila_clk = target.str_cfg.emu_clk.name

        props = {}

        # set the number samples per signal
        props['CONFIG.C_DATA_DEPTH'] = str(depth)

        # add a pipelined input to reduce burden on timing closure
        props['CONFIG.C_INPUT_PIPE_STAGES'] = '1'

        # enable capture control
        props['CONFIG.C_EN_STRG_QUAL'] = 'true'

        # two comparators per probe are recommended when using capture control (per UG908, p 40)
        props['CONFIG.ALL_PROBE_SAME_MU_CNT'] = '2'

        # specify all signals to be probed
        signals = (scfg.digital_probes + scfg.analog_probes + [target.str_cfg.time_probe] +
                  [target.str_cfg.dec_cmp])
        print(f"Signals: {[f'{signal.name}' for signal in signals]}")

        # Set number of probes in total
        props['CONFIG.C_NUM_OF_PROBES'] = str(len(signals))

        for k, signal in enumerate(signals):
            # Add depth for ila signal
            props[f'CONFIG.C_PROBE{k}_WIDTH'] = str(signal.width)

        super().__init__(ip_name='ila', props=props, ip_dir=target.ip_dir)


def main():
    print(TemplILA(target=FPGATarget(prj_cfg=EmuConfig(root='test', cfg_file=''))).render())

if __name__ == "__main__":
    main()